-- NOTE: To open F10 Menu on mission start, add the following trigger actions after loading this script
-- 1. This opens the radio menu (Given command number works on P-51 and may need changed if spawning in a different aircraft)
-- Action:
--		X: SET COMMAND
-- Command:
-- 		1591
--
-- 2. This select the F10 option to open that menu (Given command number works on P-51 and may need changed if spawning in a different aircraft)
-- Action:
--		X: SET COMMAND
-- Command:
-- 		975
--
local f10Menu = {}
local logData = {}
local validDirections = {
    ['12 O\'Clock'] = 0,
    ['3 O\'Clock'] = 90,
    ['6 O\'Clock'] = 180,
    ['9 O\'Clock'] = 270,
    ['Random'] = 'random'
}
local directionKeys = {'12 O\'Clock', '3 O\'Clock', '6 O\'Clock', '9 O\'Clock', 'Random'}
local validDistances = {
    ['1 Mile'] = 1,
    ['3 Miles'] = 3,
    ['5 Miles'] = 5,
    ['10 Miles'] = 10
}
local distanceKeys = {'1 Mile', '3 Miles', '5 Miles', '10 Miles'}
local validHeightDifferences = {
	['-5000 Feet'] = -5000,
	['-1000 Feet'] = -1000,
	['-500 Feet'] = -500,
	['-100 Feet'] = -100,
	['Same Height'] = 0,
	['100 Feet'] = 100,
	['500 Feet'] = 500,
	['1000 Feet'] = 1000,
	['5000 Feet'] = 5000
}
local heightDifferenceKeys = {'-5000 Feet', '-1000 Feet', '-500 Feet', '-100 Feet', 'Same Height', '100 Feet', '500 Feet', '1000 Feet', '5000 Feet'}
local validSpeedDifferences = {
    ['-100 Knots'] = -100,
    ['-50 Knots'] = -50,
    ['Same Speed'] = 0,
    ['50 Knots'] = 50,
    ['100 Knots'] = 100
}
local speedDifferenceKeys = {'-100 Knots', '-50 Knots', 'Same Speed', '50 Knots', '100 Knots'}
local validSkills = {'Rookie', 'Trained', 'Veteran', 'Ace'}
local engagementSettings = {
    scenario = 'ACM',
    direction = validDirections['12 O\'Clock'], -- 12 o clock
	heightDifference = 0, -- feet
    miles = 3,
    speedDifference = 0 -- mph
}

-- Maximum OpFor AI spawned aircraft allowed at one time
glbMaxOpForSpawnable = 24

validAiData = {}
glbUnitId = 4000
glbGroupId = 4000
tblBlueCallSignsOpFor = {"Enfield", "Springfield", "Uzi", "Colt", "Dodge", "Ford", "Chevy", "Pontiac"}
glbCurrentBlueCallSignOpFor = 1
glbOnBoardNumOpFor = 200

glbTblOpForSpawned = {}
glbOpForSpawned = 0

-- [[ Red WW2 & Korea era fighters ]]
table.insert(validAiData, {
    unitType = "FW-190A8",
    country = country.id.CJTF_RED,
    fuel = 409,
    flares = 0,
    chaff = 0,
    livery = "Factory skin",
    category = Group.Category.AIRPLANE,
    coalition = coalition.side.RED,
    loadouts = {
        ACM = {}
    }
})

table.insert(validAiData, {
    unitType = "FW-190D9",
    country = country.id.CJTF_RED,
    fuel = 388,
    flares = 0,
    chaff = 0,
    livery = "FW-190D9_13.JG 51_Heinz Marquardt",
    category = Group.Category.AIRPLANE,
    coalition = coalition.side.RED,
    loadouts = {
        ACM = {}
    }
})

table.insert(validAiData, {
    unitType = "Bf-109K-4",
    country = country.id.CJTF_RED,
    fuel = 296,
    flares = 0,
    chaff = 0,
    livery = "Bf-109 K4 Jagdgeschwader 53",
    category = Group.Category.AIRPLANE,
    coalition = coalition.side.RED,
    loadouts = {
        ACM = {}
    }
})

-- [[ MIST functions ]]

--- Creates a deep copy of a object.
-- Usually this object is a table.
-- See also: from http://lua-users.org/wiki/CopyTable
-- @param object object to copy
-- @return copy of object
local function deepCopy(object)
    local lookup_table = {}
    local function _copy(object)
        if type(object) ~= "table" then
            return object
        elseif lookup_table[object] then
            return lookup_table[object]
        end
        local new_table = {}
        lookup_table[object] = new_table
        for index, value in pairs(object) do
            new_table[_copy(index)] = _copy(value)
        end
        return setmetatable(new_table, getmetatable(object))
    end
    return _copy(object)
end

local function feetToMeters(feet)
    return feet * 0.3048
end

--- Converts knots to meters per second.
-- @param knots speed in knots
-- @return speed in m/sQ
local function knotsToMps(knots)
    return knots * 1852 / 3600
end

--- Converts meters per second to knots.
-- @param mps speed in m/s
-- @return speed in knots
local function mpsToKnots(mps)
    return mps * 3600 / 1852
end

--- Converts nautical miles to meters.
-- @param nm distance in nautical miles
-- @return distance in meters
local function NMToMeters(nm)
    return nm * 1852
end

--- Simple rounding function.
-- From http://lua-users.org/wiki/SimpleRound
-- use negative idp for rounding ahead of decimal place, positive for rounding after decimal place
-- @tparam number num number to round
-- @param idp
local function round(num, idp)
    local mult = 10 ^ (idp or 0)
    return math.floor(num * mult + 0.5) / mult
end

local function projectPoint(point, dist, theta)
    local newPoint = {}
    if point.z then
        newPoint.z = round(math.sin(theta) * dist + point.z, 3)
        newPoint.y = deepCopy(point.y)
    else
        newPoint.y = round(math.sin(theta) * dist + point.y, 3)
    end
    newPoint.x = round(math.cos(theta) * dist + point.x, 3)

    return newPoint
end
-- [[ End MIST Functions]]

local function spawnAi(playerUnitName, aiDetails)
    local player = Unit.getByName(playerUnitName)
    -- local headingRad = mist.getHeading(player, true)
    local playerPos = player:getPosition()
    local headingRad = math.atan2(playerPos.x.z, playerPos.x.x)
    if headingRad < 0 then
        headingRad = headingRad + 2 * math.pi -- put heading in range of 0 to 2*pi
    end
    -- local playerSpeed = mist.vec.mag(player:getVelocity())
    local playerVelocity = player:getVelocity()
    local playerSpeed = (playerVelocity.x ^ 2 + playerVelocity.y ^ 2 + playerVelocity.z ^ 2) ^ 0.5
    local aiSpeed = playerSpeed + knotsToMps(engagementSettings.speedDifference)
    if aiSpeed == 0 then
        -- aiSpeed = mist.utils.knotsToMps(250)
        aiSpeed = knotsToMps(250)
    end
    local headingDeg = math.deg(headingRad)

    local relativeDeg = engagementSettings.direction
    if (engagementSettings.direction == 'random') then
        relativeDeg = math.random(1, 360)
    end

    local newHeadingDeg = headingDeg + relativeDeg
	if newHeadingDeg > 360 then
		newHeadingDeg = newHeadingDeg - 360
	end
	local newHeadingRad = math.rad(newHeadingDeg)
    local aiHeadingRad = newHeadingRad + math.pi
    if newHeadingRad >= 2 * math.pi then
        aiHeadingRad = newHeadingRad - math.pi
    end
    local playerPoint = player:getPoint()
    local wp1Point = projectPoint(playerPoint, NMToMeters(engagementSettings.miles), newHeadingRad)
	wp1Point.y = wp1Point.y + feetToMeters(engagementSettings.heightDifference)
	local groundHeight = land.getHeight({x = wp1Point.x, y = wp1Point.z})
	-- Make sure ai is at least 100 feet above ground
	if wp1Point.y - groundHeight - 30.48 < 0 then
		wp1Point.y = groundHeight + 30.48
	end
    local wp2Point = projectPoint(playerPoint, 1, newHeadingRad)

    glbUnitId = glbUnitId + 1
    glbGroupId = glbGroupId + 1
    glbOnBoardNumOpFor = glbOnBoardNumOpFor + 1
    if aiDetails.fuel == nil then
        aiDetails.fuel = 2000
    end
    if aiDetails.flares == nil then
        aiDetails.flares = 32
    end
    if aiDetails.chaff == nil then
        aiDetails.chaff = 32
    end
    if aiDetails.livery == nil then
        aiDetails.livery = ""
    end
    if aiDetails.category == nil then
        aiDetails.category = Group.Category.AIRPLANE
    end
    if aiDetails.coalition == nil then
        aiDetails.coalition = coalition.side.NEUTRAL
    end
    aiDetails.pylons = {}
    if aiDetails.loadouts[engagementSettings.scenario] ~= nil then
        aiDetails.pylons = aiDetails.loadouts[engagementSettings.scenario]
    end
    aiDetails.name = tostring(aiDetails.unitType) .. "-" .. tostring(glbUnitId)
    aiDetails.unitId = glbUnitId
    aiDetails.groupId = glbGroupId
    aiDetails.callsign = tostring(glbOnBoardNumOpFor)

    if (coalition.getCountryCoalition(aiDetails.country) == coalition.side.BLUE) then
        aiDetails.callsign = tblBlueCallSignsOpFor[glbCurrentBlueCallSignOpFor] .. "11"
        glbCurrentBlueCallSignOpFor = glbCurrentBlueCallSignOpFor + 1
        if (glbCurrentBlueCallSignOpFor > #tblBlueCallSignsOpFor) then
            glbCurrentBlueCallSignOpFor = 1
        end
    end
    local aiGroup = {
        ["lateActivation"] = false,
        ["tasks"] = {}, -- end of ["tasks"]
        ["radioSet"] = false,
        ["task"] = "CAP",
        ["uncontrolled"] = false,
        ["route"] = {
            ["routeRelativeTOT"] = false,
            ["points"] = {
                [1] = {
                    ["alt"] = wp1Point.y, -- altitude
                    ["action"] = "Turning Point",
                    ["alt_type"] = "BARO",
                    ["speed"] = aiSpeed,
                    ["task"] = {
                        ["id"] = "ComboTask",
                        ["params"] = {
                            ["tasks"] = {
                                [1] = {
                                    ["enabled"] = true,
                                    ["key"] = "CAP",
                                    ["id"] = "EngageTargets",
                                    ["number"] = 1,
                                    ["auto"] = true,
                                    ["params"] = {
                                        ["targetTypes"] = {
                                            [1] = "Air"
                                        }, -- end of ["targetTypes"]
                                        ["priority"] = 0
                                    } -- end of ["params"]
                                }, -- end of [1]
                                [2] = {
                                    ["enabled"] = true,
                                    ["auto"] = true,
                                    ["id"] = "WrappedAction",
                                    ["number"] = 2,
                                    ["params"] = {
                                        ["action"] = {
                                            ["id"] = "Option",
                                            ["params"] = {
                                                ["value"] = true,
                                                ["name"] = 17
                                            } -- end of ["params"]
                                        } -- end of ["action"]
                                    } -- end of ["params"]
                                }, -- end of [2]
                                [3] = {
                                    ["enabled"] = true,
                                    ["auto"] = true,
                                    ["id"] = "WrappedAction",
                                    ["number"] = 3,
                                    ["params"] = {
                                        ["action"] = {
                                            ["id"] = "Option",
                                            ["params"] = {
                                                ["value"] = 4,
                                                ["name"] = 18
                                            } -- end of ["params"]
                                        } -- end of ["action"]
                                    } -- end of ["params"]
                                }, -- end of [3]
                                [4] = {
                                    ["enabled"] = true,
                                    ["auto"] = true,
                                    ["id"] = "WrappedAction",
                                    ["number"] = 4,
                                    ["params"] = {
                                        ["action"] = {
                                            ["id"] = "Option",
                                            ["params"] = {
                                                ["value"] = true,
                                                ["name"] = 19
                                            } -- end of ["params"]
                                        } -- end of ["action"]
                                    } -- end of ["params"]
                                }, -- end of [4]
                                [5] = {
                                    ["enabled"] = true,
                                    ["auto"] = true,
                                    ["id"] = "WrappedAction",
                                    ["number"] = 5,
                                    ["params"] = {
                                        ["action"] = {
                                            ["id"] = "Option",
                                            ["params"] = {
                                                ["targetTypes"] = {}, -- end of ["targetTypes"]
                                                ["name"] = 21,
                                                ["value"] = "none;",
                                                ["noTargetTypes"] = {
                                                    [1] = "Fighters",
                                                    [2] = "Multirole fighters",
                                                    [3] = "Bombers",
                                                    [4] = "Helicopters",
                                                    [5] = "Infantry",
                                                    [6] = "Fortifications",
                                                    [7] = "Tanks",
                                                    [8] = "IFV",
                                                    [9] = "APC",
                                                    [10] = "Artillery",
                                                    [11] = "Unarmed vehicles",
                                                    [12] = "AAA",
                                                    [13] = "SR SAM",
                                                    [14] = "MR SAM",
                                                    [15] = "LR SAM",
                                                    [16] = "Aircraft Carriers",
                                                    [17] = "Cruisers",
                                                    [18] = "Destroyers",
                                                    [19] = "Frigates",
                                                    [20] = "Corvettes",
                                                    [21] = "Light armed ships",
                                                    [22] = "Unarmed ships",
                                                    [23] = "Submarines",
                                                    [24] = "Cruise missiles",
                                                    [25] = "Antiship Missiles",
                                                    [26] = "AA Missiles",
                                                    [27] = "AG Missiles",
                                                    [28] = "SA Missiles"
                                                } -- end of ["noTargetTypes"]
                                            } -- end of ["params"]
                                        } -- end of ["action"]
                                    } -- end of ["params"]
                                } -- end of [5]
                            } -- end of ["tasks"]
                        } -- end of ["params"]
                    }, -- end of ["task"]
                    ["type"] = "Turning Point",
                    ["ETA"] = 0,
                    ["ETA_locked"] = true,
                    ["y"] = wp1Point.z, -- latitude
                    ["x"] = wp1Point.x, -- longitude
                    ["formation_template"] = "",
                    ["speed_locked"] = true
                }, -- end of [1]
                [2] = {
                    ["alt"] = wp2Point.y, -- altitude
                    ["action"] = "Turning Point",
                    ["alt_type"] = "BARO",
                    ["speed"] = aiSpeed,
                    ["task"] = {
                        ["id"] = "ComboTask",
                        ["params"] = {
                            ["tasks"] = {} -- end of ["tasks"]
                        } -- end of ["params"]
                    }, -- end of ["task"]
                    ["type"] = "Turning Point",
                    ["ETA"] = 601.84765346067,
                    ["ETA_locked"] = false,
                    ["y"] = wp2Point.z, -- latitude
                    ["x"] = wp2Point.x, -- longitude
                    ["formation_template"] = "",
                    ["speed_locked"] = true
                } -- end of [2]
            } -- end of ["points"]
        }, -- end of ["route"]
        ["groupId"] = aiDetails.groupId,
        ["hidden"] = "false",
        ["units"] = {
            [1] = {
                ["alt"] = wp1Point.y, -- altitude
                ["alt_type"] = "RADIO",
                ["hardpoint_racks"] = true,
                ["alt_type"] = "BARO",
                ["livery_id"] = aiDetails.livery,
                ["skill"] = aiDetails.skill,
                ["speed"] = aiSpeed,
                ["AddPropAircraft"] = {}, -- end of ["AddPropAircraft"]
                ["type"] = aiDetails.unitType,
                ["unitId"] = aiDetails.unitId,
                ["psi"] = -aiHeadingRad,
                ["y"] = wp1Point.z, -- latitude
                ["x"] = wp1Point.x, -- longitude
                ["name"] = aiDetails.name,
                ["payload"] = {
                    ["pylons"] = aiDetails.pylons,
                    ["fuel"] = aiDetails.fuel,
                    ["flare"] = aiDetails.flares,
                    ["chaff"] = aiDetails.chaff,
                    ["gun"] = 100
                }, -- end of ["payload"]
                ["heading"] = aiHeadingRad,
                ["callsign"] = {
                    [1] = 1,
                    [2] = 1,
                    ["name"] = aiDetails.callsign,
                    [3] = 1
                }, -- end of ["callsign"]
                ["onboard_num"] = aiDetails.callsign
            } -- end of [1]
        }, -- end of ["units"]
        ["y"] = wp1Point.z, -- latitude
        ["x"] = wp1Point.x, -- longitude
        ["name"] = aiDetails.name,
        ["communication"] = true,
        ["start_time"] = 0,
        ["modulation"] = 0,
        ["frequency"] = 127.5
    }

    if ((glbOpForSpawned + #aiGroup.units) <= glbMaxOpForSpawnable) then
        for i = 1, #aiGroup.units do
            table.insert(glbTblOpForSpawned, {
                gID = aiGroup.groupId,
                uID = aiGroup.units[i].unitId,
                uName = aiGroup.units[i].name
            })
        end
        -- mist.debug.writeData(mist.utils.serialize,{'log', logData}, 'wwII.log')
        return coalition.addGroup(aiDetails.country, aiDetails.category, aiGroup)
    else
        trigger.action.outText("Cannot spawn any more OpFor aircraft, exceeds the maximum active allowed: " ..
                                   glbMaxOpForSpawnable, 10)
        trigger.action.outText("Retry after some active OpFor aircraft are destroyed or land.", 10)
        return
    end
end

for i = 1, #validAiData do
    if validAiData[i].loadouts.ACM ~= nil then
        f10Menu[validAiData[i].unitType] = missionCommands.addSubMenu(validAiData[i].unitType:gsub('-', ' '))
        for j = 1, #validSkills do
            local aiData = validAiData[i]
            aiData.skill = validSkills[j]
            f10Menu[validAiData[i].unitType .. '.' .. validSkills[j]] =
                missionCommands.addCommand(validSkills[j], f10Menu[validAiData[i].unitType], spawnAi, 'Player', aiData)
        end
    end
end

local function setDirection(direction)
    engagementSettings.direction = direction
end

local function setDistance(miles)
    engagementSettings.miles = miles
end

local function setHeightDifference(heightDifference)
    engagementSettings.heightDifference = heightDifference
end

local function setSpeedDifference(speedDifference)
    engagementSettings.speedDifference = speedDifference
end

f10Menu['settings'] = missionCommands.addSubMenu('Settings')
f10Menu['settings.direction'] = missionCommands.addSubMenu('Direction', f10Menu['settings'])
for i = 1, #directionKeys do
	local descripton = directionKeys[i]
	local direction = validDirections[descripton]
    f10Menu['settings.direction.' .. descripton] = missionCommands.addCommand(descripton, f10Menu['settings.direction'], setDirection, direction)
end
f10Menu['settings.distance'] = missionCommands.addSubMenu('Distance', f10Menu['settings'])
for i = 1, #distanceKeys do
	local descripton = distanceKeys[i]
	local distance = validDistances[descripton]
    f10Menu['settings.distance.' .. descripton] = missionCommands.addCommand(descripton, f10Menu['settings.distance'], setDistance, distance)
end
f10Menu['settings.heightDifference'] = missionCommands.addSubMenu('Height Difference', f10Menu['settings'])
for i = 1, #heightDifferenceKeys do
	local descripton = heightDifferenceKeys[i]
	local heightDifference = validHeightDifferences[descripton]
    f10Menu['settings.heightDifference.' .. descripton] = missionCommands.addCommand(descripton, f10Menu['settings.heightDifference'], setHeightDifference, heightDifference)
end
f10Menu['settings.speedDifference'] = missionCommands.addSubMenu('Speed Difference', f10Menu['settings'])
for i = 1, #speedDifferenceKeys do
	local descripton = speedDifferenceKeys[i]
	local speedDifference = validSpeedDifferences[descripton]
    f10Menu['settings.speedDifference.' .. descripton] = missionCommands.addCommand(descripton, f10Menu['settings.speedDifference'], setSpeedDifference, speedDifference)
end
-- logData['f10Menu'] = f10Menu
-- mist.debug.writeData(mist.utils.serialize, {'log', logData}, 'wwII.log')
